# DepthAI Gate
# Initial implementation in Python
# To be either ported or native modules added for interraction with USB

from flask import Flask, json, jsonify
from flask_restful import Api, Resource, reqparse, fields, marshal, inputs
import werkzeug
from depthai_gate.exceptions import PublicError
from depthai_gate.session_manager import SessionManager, Session
from depthai_gate.fwp_manager import FwpManager
from depthai_gate.discovery import DeviceDiscovery, State
import logging

DEPTHAI_GATE_HTTP_PORT = 11492
DEPTHAI_GATE_HTTPS_PORT = 11492

class CustomApi(Api):
    """
    Patch Flask-style custom error handling into the Flask-RESTful api class.
    """
    def __init__(self, *args, **kwargs):
        super(CustomApi, self).__init__(*args, **kwargs)
        self._errorhandlers = []

    def errorhandler(self, exception_type):
        """
        Defined handlers for exceptions. Example:
        @api.errorhandler(ServerError):
        def handle_server_error(error):
            response = flask.jsonify({'message': error.message})
            response.status_code = error.status_code
            return response
        """
        def wrapper(func):
            import functools
            self._errorhandlers.append((exception_type, func))
            # Sort error handlers to have sub exceptions first, so that those
            # take preference over base exceptions.

            self._errorhandlers = sorted(
                self._errorhandlers,
                key=functools.cmp_to_key(self._inheritance_comparator))
            print(self._errorhandlers)
            return func
        return wrapper

    def handle_error(self, error, previous_errors=None):
        # Keep track of previous errors in the current chain of exception
        # handling in order to prevent infinite cycles that would occur if two
        # error handler raise the exception handled by the other.
        print(f'Exception: {error}')
        previous_errors = previous_errors or []
        previous_errors.append(type(error))
        # Try to find the first custom handler for the occured exception.
        for exception_type, handler in self._errorhandlers:
            if not isinstance(error, exception_type):
                continue
            try:
                return handler(error)
            except Exception as new_error:
                if type(new_error) not in previous_errors:
                    return self.handle_error(new_error, previous_errors)
            break
        # If no matching handler was found or an infinite cycle is detected,
        # fall back to Flask-RESTful's error handling.
        return super(CustomApi, self).handle_error(error)

    @staticmethod
    def _inheritance_comparator(lhs, rhs):
        lhs_sub = issubclass(lhs, rhs)
        rhs_sub = issubclass(lhs, rhs)
        if lhs_sub and not rhs_sub:
            return -1
        if rhs_sub and not lhs_sub:
            return 1
        return 0

def main():
    # logging.basicConfig(encoding='utf-8', level=logging.DEBUG)
    logging.basicConfig(level=logging.DEBUG)
    logging.debug('DepthAI Gate - Debug Logging ON!')

    app = Flask(__name__)
    api = CustomApi(app)

    @api.errorhandler(PublicError)
    def handle_exception(err):
        """Return JSON instead of HTML for MyCustomError errors."""
        response = {
            "error": str(err),
        }
        return jsonify(response), err.code

    session_fields = {
        "id": fields.String,
        "name": fields.String,
        "fwp_checksum": fields.String,
        "fwp_version": fields.String,
        "library_version": fields.String,
        # "environment": fields.String,
        "connection_information": fields.String,
        "owner": fields.String,
        "protected": fields.Boolean,
        "create_datetime": fields.DateTime('iso8601'),
        "start_datetime": fields.DateTime('iso8601'),
        "stop_datetime": fields.DateTime('iso8601'),
        "destroy_datetime": fields.DateTime('iso8601'),
        "fwp_exists": fields.Boolean,
        "state": fields.String,
    }

    fwp_manager = FwpManager()
    session_manager = SessionManager(fwp_manager)

    # /sessions
    class SessionListAPI(Resource):
        def __init__(self, manager, fwp_manager) -> None:
            super().__init__()
            self.manager = manager
            self.fwp_manager = fwp_manager
            self.parse_create_session = reqparse.RequestParser()
            self.parse_create_session.add_argument('name', type=str, required=False, location='json')
            self.parse_create_session.add_argument('fwp_checksum', type=str, required=False, location='json')
            self.parse_create_session.add_argument('fwp_version', type=str, required=False, location='json')
            self.parse_create_session.add_argument('library_version', type=str, required=True, help='No library version specified', location='json')
            self.parse_create_session.add_argument('environment', type=str, required=False, location='json')
            self.parse_create_session.add_argument('protected', type=bool, required=False, location='json')

        def get(self):
            return {'sessions': [marshal(session, session_fields) for session in self.manager.get_all()]}

        def post(self):
            args = self.parse_create_session.parse_args()
            session = self.manager.create(args['name'], args['fwp_checksum'], args['fwp_version'], args['library_version'], args['environment'], '', '', args['protected'])
            return marshal(session, session_fields)

    # /sessions/{id}
    class SessionAPI(Resource):
        def __init__(self, manager) -> None:
            super().__init__()
            self.manager = manager
        def get(self, id):
            print(type(id))
            return marshal(self.manager.get(id), session_fields)
        def delete(self, id):
            self.manager.delete(id)
            return {'result': True}

    # /sessions/{id}/fwp
    class SessionFwpAPI(Resource):
        def __init__(self, manager, fwp_manager) -> None:
            super().__init__()
            self.manager = manager
            self.fwp_manager = fwp_manager
            self.upload_parser = reqparse.RequestParser()
            # self.upload_parser.add_argument('', type=bytes, required=True, location='data')
            self.upload_parser.add_argument('dataUrl', type=inputs.boolean, default=False)
        def post(self, id):
            parse = reqparse.RequestParser()
            parse.add_argument('file', type=werkzeug.datastructures.FileStorage, location='files')
            args = parse.parse_args()
            fwp_filestorage = args['file']
            fwp_binary = fwp_filestorage.read()
            print(len(fwp_binary))
            self.manager.get(id).fwp_checksum = self.fwp_manager.create(fwp_binary)
            return {'result': True}

    # /sessions/{id}/start
    class SessionStartAPI(Resource):
        def __init__(self, manager, fwp_manager) -> None:
            super().__init__()
            self.manager = manager
            self.fwp_manager = fwp_manager
        def post(self, id):
            self.manager.start(id)
            return {'result': True}

    # /sessions/{id}/start
    class SessionStopAPI(Resource):
        def __init__(self, manager, fwp_manager) -> None:
            super().__init__()
            self.manager = manager
            self.fwp_manager = fwp_manager
        def post(self, id):
            self.manager.stop(id)
            return {'result': True}

    # /status
    class StatusAPI(Resource):
        def get(self):
            return {'status': True}


    api.add_resource(SessionListAPI, '/api/v1/sessions', resource_class_kwargs={'manager': session_manager, 'fwp_manager': fwp_manager})
    api.add_resource(SessionAPI, '/api/v1/sessions/<string:id>', resource_class_kwargs={'manager': session_manager})
    api.add_resource(SessionFwpAPI, '/api/v1/sessions/<string:id>/fwp', resource_class_kwargs={'manager': session_manager, 'fwp_manager': fwp_manager})
    api.add_resource(SessionStartAPI, '/api/v1/sessions/<string:id>/start', resource_class_kwargs={'manager': session_manager, 'fwp_manager': fwp_manager})
    api.add_resource(SessionStopAPI, '/api/v1/sessions/<string:id>/stop', resource_class_kwargs={'manager': session_manager, 'fwp_manager': fwp_manager})
    api.add_resource(StatusAPI, '/api/v1/status')

    # retrieve serial
    # /sys/firmware/devicetree/base/serial-number
    # MX: 18443010818C0E0900
    # KB: 2b528816838860c6
    # KB: 2b08a816884861c2
    serial = 'xlink-gate'
    try:
        with open('/sys/firmware/devicetree/base/serial-number') as f:
            serial = (f.readlines()[0]).upper()
    except Exception as ex:
        logging.error(f"Couldn't retrieve devices serial number. Using default 'xlink-gate'")


    def get_state_cb():
        active_session = False
        sessions = session_manager.get_all()
        for session in sessions:
            if session.state not in ['STOPPED', 'CRASHED', 'DESTROYED']:
                active_session = True
        if active_session:
            return State.GATE_BOOTED
        return State.GATE

    discovery = DeviceDiscovery(serial, DEPTHAI_GATE_HTTP_PORT, DEPTHAI_GATE_HTTPS_PORT)
    discovery.set_state_cb(get_state_cb)
    # app.run(port=DEPTHAI_GATE_HTTPS_PORT, ssl_context='adhoc')
    app.run(host='0.0.0.0', port=DEPTHAI_GATE_HTTP_PORT)

if __name__ == '__main__':
    main()