from depthai_gate.fwp_manager import FwpManager
from depthai_gate.exceptions import PublicError
from depthai_gate.watchdog import Watchdog

from datetime import datetime
import subprocess
from flask_restful import Resource
import json
import logging
from threading import RLock
import os
import importlib.util
import sys


class Session:
    id: str
    name: str
    fwp_checksum: str
    fwp_version: str
    library_version: str
    environment: dict
    connection_information: str
    owner: str
    protected: bool
    proc: subprocess.Popen
    create_datetime: datetime
    start_datetime: datetime
    stop_datetime: datetime
    destroy_datetime: datetime
    fwp_exists: bool
    watchdog: Watchdog
    state: str # CREATED, RUNNING, STOPPED, STOPPING, CRASHED, DESTROYED


    def __init__(self, id, name, fwp_checksum, fwp_version, library_version, environment, connection_information, create_datetime, start_datetime, stop_datetime, destroy_datetime, owner, protected):
        self.id = id
        self.name = name
        self.fwp_checksum = fwp_checksum
        self.fwp_version = fwp_version
        self.library_version = library_version
        self.environment = environment
        self.connection_information = connection_information
        self.create_datetime = create_datetime
        self.start_datetime = start_datetime
        self.stop_datetime = stop_datetime
        self.destroy_datetime = destroy_datetime
        self.owner = owner
        self.protected = protected
        self.proc = None
        self.fwp_exists = False
        self.watchdog = Watchdog(id)
        self.state = 'CREATED'

class SessionManager():
    last_unique_id: int
    fwp_manager: FwpManager
    mutex: RLock
    on_event_module = None
    # session_map: Dict[str, Session]

    def load_module(self, module):

        spec = importlib.util.spec_from_file_location("module.name", str(module))
        m = importlib.util.module_from_spec(spec)
        sys.modules["module.name"] = m
        spec.loader.exec_module(m)

        return m


    def __init__(self, fwp_manager) -> None:
        super().__init__()
        self.last_unique_id = 0
        self.session_map = dict()
        self.fwp_manager = fwp_manager
        self.mutex = RLock()

    def _get_unique_session_id(self):
        with self.mutex:
            self.last_unique_id = self.last_unique_id + 1
            return str(self.last_unique_id)

    def create(self, name, fwp_checksum, fwp_version, library_version, environment, connection_information, owner, protected):
        with self.mutex:
            session = Session(self._get_unique_session_id(), name, fwp_checksum, fwp_version, library_version, environment, connection_information, datetime.now(), None, None, None, owner, protected)
            self.session_map[session.id] = session
            self.fwp_exists = self.fwp_manager.exists(session.fwp_checksum)
            return session

    def start(self, id):
        with self.mutex:
            if id not in self.session_map:
                raise PublicError(f'Session with given id: {id} does not exist')
            session = self.session_map[id]

            session.start_datetime = datetime.now()

            entry = self.fwp_manager.get_entry_point(session.fwp_checksum)
            root_folder = self.fwp_manager.get_root_folder(session.fwp_checksum)

            script_folder = root_folder / "scripts"
            custom_script = script_folder / "custom_scripts.py"

            try:
                on_event_module = self.load_module(str(custom_script))
            except Exception as e:
                print(e)
                on_event_module = None

            if on_event_module:
                try:
                    on_event_module.on_start()
                except Exception as e:
                    print(e)
 

            has_vpu_fw = self.fwp_manager.has_vpu_fw(session.fwp_checksum)
            vpu_fw = None
            if has_vpu_fw:
                vpu_fw = self.fwp_manager.get_vpu_fw(session.fwp_checksum)
                logging.debug(f'[SessionManager] Starting session ({id}) - vpu entry: {vpu_fw}')
                fw_target_name = 'luxonis_vpu.bin'
                fw_target_path = '/var/lib/firmware/intel/vpux/' + fw_target_name
                try:
                    subprocess.check_call(['cp', str(vpu_fw), fw_target_path])
                    subprocess.check_call(['StartVpu', str(fw_target_name)])
                except:
                    raise PublicError(f'Session with given id: {id} failed booting VPU')

            logging.debug(f'[SessionManager] Starting session ({id}) - fwp entry: {entry}, has vpu: {has_vpu_fw}, vpu entry: {vpu_fw}, root_folder: {root_folder}')
            fd_watchdog_read, fd_watchdog_write = os.pipe()
            session.watchdog.watchdog_fd = fd_watchdog_read
            session.proc = subprocess.Popen([entry], cwd=root_folder, close_fds=True,
                                            pass_fds=[fd_watchdog_read, fd_watchdog_write],
                                            env=dict(os.environ, DEPTHAI_GATE_WATCHDOG_FD=str(fd_watchdog_write)))
            session.watchdog.proc = session.proc
            session.state = 'RUNNING'
            session.watchdog.run()

    def stop(self, id):
        with self.mutex:
            if id not in self.session_map:
                raise PublicError(f'Session with given id: {id} does not exist')
            session = self.session_map[id]
            if session.proc is None:
                raise PublicError(f'Session with given id: {id} was not started')

            session.stop_datetime = datetime.now()
            session.watchdog.stop()
            if on_event_module:
                try:
                    on_event_module.on_stop()
                except Exception as e:
                    print(e)
            session.proc.terminate()
            session.state = 'STOPPING'

    def destroy(self, id):
        with self.mutex:
            if id not in self.session_map:
                raise PublicError(f'Session with given id: {id} does not exist')
            session = self.session_map[id]

            if on_event_module:
                try:
                    on_event_module.on_destroy()
                except Exception as e:
                    print(e)

            session.watchdog.stop()
            session.proc.kill()
            session.destroy_datetime = datetime.now()
            session.state = 'DESTROYED'

    def refresh_process_state(self, id):
        with self.mutex:
            if id not in self.session_map:
                raise PublicError(f'Session with given id: {id} does not exist')
            session = self.session_map[id]
            if session.proc is not None:
                retcode = session.proc.poll()
                logging.debug(f'[SessionManager] Session ({id}) process poll: {retcode}')
                if retcode is not None:
                    session.state = 'STOPPED'
                    if retcode != 0:
                        session.state = 'CRASHED'
            else:
                logging.debug(f'[SessionManager] Session ({id}) process not started...')

    def get(self, id):
        with self.mutex:
            if id not in self.session_map:
                raise PublicError(f'Session with given id: {id} does not exist')
            self.refresh_process_state(id)
            return self.session_map[id]

    def get_all(self):
        with self.mutex:
            sessions = []
            for sessionId in self.session_map:
                sessions.append(self.get(sessionId))
            return sessions

    def delete(self, id):
        with self.mutex:
            if id not in self.session_map:
                raise PublicError(f'Session with given id: {id} does not exist')
            self.destroy(id)
            del self.session_map[id]
